{*************************************************************************}
{*                                                                       *}
{*                  III Olimpiada Informatyczna                          *}
{*                                                                       *}
{*   Rozwiazanie zadania: MOKRA ROBOTA                                   *}
{*   Plik:                KOLEJKA.PAS - modul Kolejka                    *}
{*   Autor:               Marcin Engel                                   *}
{*************************************************************************}

unit Kolejka;

{--------------------------------------------------------}
{        Modul strukture danych dla zadania              }
{                   Mokra Robota                         }
{                                                        }
{  Autor: M. Engel            Data ost. mod.: 29.11.1995 }
{--------------------------------------------------------}

interface

 uses
  Typy;

   { Ile we wszystkich procedurach oznacza liczbe naczyn. }

 procedure Inicjuj(Ile:NumerNaczynia; s:Stan; WagaKoncowa:byte);
  { Inicjuje strukture danych, wstawia do niej stan s. }

 procedure Wstaw(Ile:NumerNaczynia; s:Stan; Krok:Kroki; WagaKoncowa:byte);
  { Wstawia stan s z liczba krokow Krok do struktury, o ile jeszcze
    go tam nie bylo  }

 procedure Pobierz(Ile:NumerNaczynia; var s:Stan; var Krok:Kroki);
  { Pobiera pierwszy element z kolejki stanow wraz z odpowiednia liczba
    krokow }

 function KolejkaPusta:boolean;
  { czy kolejka jest pusta }

 procedure UsunZbedne;
  { usuwa juz nieistotne informacje }

 procedure Sprzataj;
  { wielkie sprzatanie }

implementation

{---------------------------------------------------------}
{               Opis struktury danych                     }
{                                                         }
{ Elementy wstawiane do kolejki sa pamietane w drzewie    }
{ k.Byly[w], gdzie w jest waga wstawianego stanu.         }
{ Elementy w drzewach sa polaczone za pomoce dodatkowej   }
{ listy jednokierunkowej, ktora jest realizacja kolejki.  }
{ Wstawienie el. do struktury, oznacza wstawienie go do   }
{ odpowiedniego drzewa z jednoczesnym dolaczeniem na      }
{ koniec kolejki. Jednoczesnie aktualizuje sie liczbe     }
{ elementow o danej wadze w kolejce.                      }
{ Pobranie el. z kolejki, oznacza pobranie pierwszego     }
{ elementu z kolejki, aktualizacje licznikow, ale pozos-  }
{ tawienie el. w drzewie. Element o wadze w jest tak      }
{ dlugo w drzewie, jak dlugo kolejka zawiera elementy     }
{ o wadze niemniejszej niz w. Usuwanie zbednych elementow }
{ odbywa sie w procedurze UsunZbedne                      }
{---------------------------------------------------------}


 type
  PElement=^Element;

  Element =record
            kr        :Kroki;       { liczba wykonanych krokow }
            st        :Stan;        { stan naczyn }
            Lewy,Prawy:PElement;    { wsk. do poddrzew }
            Nast      :PElement     { wsk. do nast. w kolejce }
           end;
  Kolej   =record
            Pier,Ost:PElement;      { wsk. do pierszego i ostatniego }
            MaxWag  :-1..MaxWaga;   { maksymalna waga elementu w kolejce }
            Byly    :array[Waga] of record
                                     Ile:longint;
                                      {l. elem. o wadze w w kolejce }
                                     Korz:PElement
                                      {drzewa przechowujace ogladane el}
                                    end
           end; {rekord Kolej}

 var
  k:Kolej;

 procedure Inicjuj(Ile:NumerNaczynia; s:Stan; WagaKoncowa:byte);
  var i:Waga;
 begin
  with k do begin
   Pier:=nil;
   Ost:=nil;
   for i:=0 to MaxWaga do begin
    Byly[i].Korz:=nil;
    Byly[i].Ile:=0
   end;
   Wstaw(Ile,s,0,WagaKoncowa);     { Wstawienie stanu pocz. }
   MaxWag:=LacznaPoj(Ile,s)
  end {with }
 end; {Inicjuj}

 function JuzJest(Ile:NUmerNaczynia; s:Stan; w:Waga):boolean;
   { czy stan s o wadze w jest w drzewie }
  var
  p   :PElement;
  Znal:boolean;
  Por :-1..1;
 begin
  p:=k.Byly[w].Korz;
  Znal:=false;
  while not Znal and (p <> nil) do begin
   Por:=Porownaj(Ile,p^.st,s);
   if Por=-1 then p:=p^.Lewy
   else if Por=1 then p:=p^.Prawy
        else Znal:=true
  end; {while }
  JuzJest:=Znal
 end; {JuzJest}

 procedure WstawDoKolejki(PElem:PElement; w:Waga);
  { Pomocnicza wywolywana w Wstaw. Dowiazuje element PElem^ do listy }
 begin
  with k,PElem^ do begin
   if Ost<>nil then Ost^.Nast:=PElem
   else Pier:=PElem;
   Nast:=nil;
   Ost:=PElem;
   Byly[w].Ile:=Byly[w].Ile+1
  end {with}
 end; {WstawDoKolejki}

 procedure WstawDoDrzewa(Ile:NumerNaczynia; PElem:PElement; w:Waga);
   { Pomocnicza - wywolywana przez Wstaw. Wstawia PElem^ o wadze w do drzewa. }
  var
   p,q:PElement;
   Por:-1..1;
 begin
  p:=k.Byly[w].Korz;
  q:=nil;
  while p<>nil do begin
   Por:=Porownaj(Ile,p^.st,PElem^.st);
   q:=p;
   if Por=-1 then p:=p^.Lewy
   else p:=p^.Prawy
  end; {while }
  with PElem^ do begin
   Lewy:=nil;
   Prawy:=nil;
   if q=nil then k.Byly[w].Korz:=PElem   { drzewo puste }
   else                 { niepuste }
    if Por=1 then q^.Prawy:=PElem
    else q^.Lewy:=PElem
  end {with }
 end; {WstawDoDrzewa}

 procedure Wstaw(Ile:NumerNaczynia; s:Stan; Krok:Kroki; WagaKoncowa:byte);
  var
   w    :Waga;
   PElem:PElement;
 begin
  w:=LacznaPoj(Ile,s);        { waga stanu s }
  if not JuzJest(Ile,s,w) and (w>=WagaKoncowa) then begin
   new(PElem);
   with PElem^ do begin
    kr:=Krok;
    st:=s
   end; {with }
   WstawDoDrzewa(Ile,PElem,w);
   WstawDoKolejki(PElem,w)
  end {if }
 end; {Wstaw}

 function KolejkaPusta:boolean;
 begin
  KolejkaPusta:=k.Pier=nil
 end; {KolejkaPusta}

 procedure Pobierz(Ile:NumerNaczynia; var s:Stan; var Krok:Kroki);
  var w:Waga;
 begin
  with k.Pier^ do begin
   s:=st;
   Krok:=kr;
   k.Pier:=Nast;
   Nast:=nil
  end; {with}
  if k.Pier=nil then k.Ost:=nil;
  w:=LacznaPoj(Ile,s);
  k.Byly[w].Ile:=k.Byly[w].Ile-1 { aktualizacja licznikow }
 end; {Pobierz}

 procedure ZwolnijDrzewo(p:PElement);
 begin
  if p<>nil then begin
   ZwolnijDrzewo(p^.Lewy);
   ZwolnijDrzewo(p^.Prawy);
   Dispose(p);
   p:=nil
  end {if }
 end; {ZwolnijDrzewo}

 procedure UsunZbedne;
 begin
  with k do
   while (MaxWag>=0) and (Byly[MaxWag].Ile=0) do begin
    ZwolnijDrzewo(Byly[MaxWag].Korz);
    MaxWag:=MaxWag-1
   end {while}
 end; {UsunZbedne}

 procedure Sprzataj;
 begin
  with k do begin
   while MaxWag>=0 do begin
    ZwolnijDrzewo(Byly[MaxWag].Korz);
    Byly[MaxWag].Ile:=0;
    MaxWag:=MaxWag-1
   end; {while}
   Pier:=nil;
   Ost:=nil
  end {with }
 end; {Sprzataj}

end. {unit Kolejka}
