{*************************************************************************}
{*                                                                       *}
{*                  III Olimpiada Informatyczna                          *}
{*                                                                       *}
{*   Rozwiazanie zadania: MOKRA ROBOTA                                   *}
{*   Plik:                WSTEPNE.PAS - modul Wstepne                    *}
{*   Autor:               Marcin Engel i Piotr Chrzastowski-Wachtel      *}
{*************************************************************************}

unit Wstepne;

{--------------------------------------------------------}
{    Modul definiujacy wstepne operacje sprawdzajace     }
{ osiagalnosc stanu koncowego z poczatkowego dla zadania }
{                  Mokra Robota                          }
{                                                        }
{  Autor: M. Engel            Data ost. mod.: 29.11.1995 }
{         P. Chrzastowski-Wachtel             20.07.1996 }
{--------------------------------------------------------}

interface
 uses Typy;

 function AnalizaWstepna(Ile:NumerNaczynia; sp,sk:Stan):boolean;
  { Funkcja dokonuje analizy wstepnej zadania, sprawdzajac proste }
  { warunki konieczne na osiagalnosc konfiguracji koncowej.       }
  { Ma wartosc false, jesli z konfiguracji poczatkowej nie mozna  }
  { przejsc do koncowej.                                          }

 function AnalizaWTrakcie(Ile:NumerNaczynia; sp,sk:Stan):boolean;

implementation

 function NWD(a,b:Pojemnosc):Pojemnosc;
   { Najwiekszy Wspolny Dzielnik liczb a i b.     }
   { W wywolaniu max(a,b)>0 i min(a,b)>=0         }
  var
   pom,x,y:Pojemnosc;
 begin
  if a>=b then begin x:=a;  y:=b end
  else begin x:=b;  y:=a end;
   { 0<>x=max(a,b)>=min(a,b)=y>=0 }
  while y>0 do begin
   { Niezmiennik: x>=y>=0 and NWD(a,b)=NWD(x,y)      }
   { Funkcja ograniczajaca: f(x,y)=y, bo (x mod y)<y }
   pom:=y;
   y:=x mod y; { poprawne, bo y>0 }
   x:=pom
  end; {while}
   { NWD(a,b)=NWD(x,0)=x }
  NWD:=x
 end; {NWD}

 function NWDStanu(Ile:NumerNaczynia; s:Stan):Pojemnosc;
   { Funkcja liczy NWD(s[1], ..., s[Ile]), pomijajac wartosci zerowe }
  var
   Dzielnik:Pojemnosc;
   k       :1..MaxNaczynie+1;
 begin
  k:=1;
  repeat
   Dzielnik:=Zawartosc(s,k);
   k:=k+1
  until (Dzielnik<>0) or (k>Ile);
  if Dzielnik=0 then Dzielnik:=1;
  while (Dzielnik>1) and (k<=Ile) do begin
   Dzielnik:=NWD(Dzielnik,Zawartosc(s,k));
   k:=k+1
  end; {while}
  NWDStanu:=Dzielnik
 end; {NWDStanu}

 function PojedynczeNaczynie(sp,sk:Stan):boolean;
  { Funkcja sprawdza osiagalnosc sk z sp w przypadku, gdy jest jedno naczynie. }
 begin
  PojedynczeNaczynie:=Rowne(1,sp,sk) or  { sp = sk  --- 0 krokow }
                      Puste(sk,1)        { sk [1] = 0 --- 1 krok }
 end; {PojedynczeNaczynie}

 function JestWystarczajacoDuzo(Ile:NumerNaczynia; sp,sk:Stan):boolean;
   { Pierwszym krokiem jest zawsze wylanie zawartosci ktoregos naczynia.
     Funkcja ma wartosc true, jesli to co pozostanie wystarczy do
     osiagniecia konfiguracji koncowej. }
  var
   x,Min:Pojemnosc;
   i    :NumerNaczynia;
 begin
  Min:=Zawartosc(sp,1);
  for i:=2 to Ile do begin
   x:=Zawartosc(sp,i);
   if x<=Min then Min:=x
  end; {for}
   { Min = min(sp[1], ..., sp [Ile]) }
  JestWystarczajacoDuzo:=LacznaPoj(Ile,sp)-Min>=LacznaPoj(Ile,sk)
 end; {JestWystarczajacoDuzo}

 function JestPusteLubPelne(Ile:NumerNaczynia; sp,sk:Stan):boolean;
  { Przyjmuje wartosc true wtedy i tylko wtedy, gdy w koncowym stanie
    przynajmniej jedno naczynie jest puste lub pelne.}
  var
   i:NumerNaczynia;
   z:Pojemnosc;
 begin
  JestPusteLubPelne:=false;
  for i:=1 to Ile do begin
   z:=Zawartosc(sk,i);
   if (z=0) or (z=Zawartosc(sp,i)) then JestPusteLubPelne:=true
  end {for}
 end; {JestPusteLubPelne}

 function JestPodzielne(Ile:NumerNaczynia; sp,sk:Stan):boolean;
  { Jesli stan sk jest osiagalny z sp, to dla kazdego 1 <= k <= Ile
    NWD(sp[1],...,sp[Ile]) dzieli sp[k]. }
  var
   k       :1..MaxNaczynie+1;
   Dzielnik:Pojemnosc;
   JestOK  :Boolean;
 begin
  Dzielnik:=NWDStanu(Ile,sp);
  if Dzielnik<>1 then begin
   k:=1;  JestOK:=true;
   while JestOK and (k<=Ile) do begin
    JestOK:=Zawartosc(sk,k) mod Dzielnik=0;
    k:=k+1
   end; {while}
   JestPodzielne:=JestOK
  end {if }
  else JestPodzielne:=true { NWD = 1, wiec jest dobrze }
 end; {JestPodzielne}

 function AnalizaWstepna(Ile:NumerNaczynia; sp,sk:Stan):boolean;
  var
   i     :integer;
   JestOK:boolean;
   x     :Pojemnosc;
 begin
  if Ile=1 then AnalizaWstepna:=PojedynczeNaczynie(sp,sk)
  else AnalizaWstepna:=JestPodzielne(Ile,sp,sk) and
                       JestWystarczajacoDuzo(Ile,sp,sk) and
                       JestPusteLubPelne(Ile,sp,sk)
 end; {AnalizaWstepna}

 function AnalizaWTrakcie(Ile:NumerNaczynia; sp,sk:Stan):boolean;
 begin
  if Ile=1 then AnalizaWTrakcie:=PojedynczeNaczynie(sp,sk)
  else AnalizaWTrakcie:=LacznaPoj(Ile,sp)>=LacznaPoj(Ile,sk)
 end; {AnalizaWTrakcie}

end. {unit Wstepne}
