{*************************************************************************}
{*                                                                       *}
{*                  III Olimpiada Informatyczna                          *}
{*                                                                       *}
{*   Rozwiazanie zadania: ZAMEK                                          *}
{*   Plik:                ZAM.PAS                                        *}
{*   Autor:               Krzysztof Diks                                 *}
{*************************************************************************}

{$M 65500,0,655360}
program Zamek;
 const
  MaksN    =5000;
   {Maksymalna liczba wierzcholkow w wielokacie-zamku.}
  MaksNDiv2=MaksN div 2;
   {Maksymalna liczba bokow poziomych (pionowych).}
  MaksXY   =10000;
   {Maksymalna wartosc wsp. x (y).}
  MaksLiczbaWezlowWDrzewiePrzekroju=8191;
   {Maksymalna liczba warstw jest rowna 2500. Stad ll<=4096, a stad
    liczba wezlow w drzewie przkroju jest <= 8191.}
 type
  WierzchTab   =array[1..2] of 0..MaksXY;
   {Wierzcholek wielokata, wsp. x - 1-szy element tablicy,
                           wsp. y - 2-gi element tablicy.}
  WielokatWsk  =^WielokatTab;
   {Wskaznik do tablicy wierzcholkow wielokata.}
  WielokatTab  =array[1..MaksN] of WierzchTab;
   {Tablica wierzcholkow wielokata.}
  WarstwyWsk   =^WarstwyTab;
  WarstwyTab   =array[1..MaksN] of integer;
   {Numer warstwy do ktorej nalezy wierzcholek.}
  BokiInfoWsk  =^BokiInfoTab;
  BokiInfoTab  =array[1..MaksNDiv2] of integer;
   {Rozne informacje zwiazane z bokami.}
  PrzekrojWsk  =^PrzekrojTab;
  PrzekrojTab  =array[0..MaksLiczbaWezlowWDrzewiePrzekroju] of 0..MaksN;
   {Przekroj}
   {Typy do budowy grafu prostokatow.}
  WskDoNastProstokataNaLiscie=^NastNaLiscie;
  NastNaLiscie =record
                 Nr  :integer;
                  {Numer prostokata w porzadku zamykania.}
                 Nast:WskDoNastProstokataNaLiscie
                end;
  ProstokatTab =record  {informacje o prostokacie}
                 Ld               : WierzchTab;  {wsp. dolnego-lewego rogu}
                 Wys,Szer         :integer;      {wysokosc, szerokosc}
                 ListaSasiednichPr:WskDoNastProstokataNaLiscie
                end;
  ProstokatyTab=array[1..MaksNDiv2] of ProstokatTab;
   {Tablica prostokatow.}

   {UWAGA: Czesc tablic uzywanych w programie jest dynamiczna.
    Jest to spowodowane mala iloscia pamieci przeznaczanej przez
    Turbo-Pascal na zmienne statyczne.}

{****************STRUKTURY DANYCH*****************************}

 var
  n,nDiv2        :integer;
   {Liczba wierzcholkow w wielokacie i polowa liczby wierzcholkow.}
  w              : WielokatWsk;
   {w^ - ciag kolejnych wierzcholkow wielokata wejsciowego. Wielokat
    zostanie zapamietany w taki sposob, aby bok w^[1]--w^[2] byl bokiem
    poziomym. Wowczas bok w^[i]--w^[i+1] jest poziomy, gdy i jest
    nieparzyste, a pionowy gdy i jest parzyste.
    Identyfikatorem boku w^[i]--w^[i+1] jest i.}
  BokiPionowe    :BokiInfoWsk;
   {Tablica bokow pionowych uporzadkowanych zgodnie z relacja <
    zdefiniowana z opisu.}
  Warstwy        :WarstwyWsk;
   {Dla kazdego wierzcholka pamietamy nr warstwy, do ktorej nalezy bok
    poziomy o koncu w tym wierzcholku.}
  lw,ll          :integer;
   {liczba warstw, liczba lisci w drzewie Przekroju}
  Otw,LewyBok    : BokiInfoTab;
   {Jesli w^[i]--w^[i+1] jest poziomym bokiem otwierajacym w aktualnym
    przekroju, to LewyBok[i div 2 + 1] jest numerem linii pionowej w siatce
    zawierajacej lewy bok otwartego (i jeszcze nie zamknietego) prostokata,
    ktorego dolny bok jest zawarty w boku w^[i]--w^[i+1] (pamietajmy, ze
    boki sa identyfikowane kolejnymi liczbami nieparzystymi).
    Otw[i div 2 + 1] jest numerem tego prostokata w kolejnosci otwierania.}
  Przekroj       :PrzekrojWsk;
   {Struktura danych reprezentujaca boki poziome w aktualnym przekroju;
    Przekroj[0] pelni role straznika i jest zawsze rowne 0.}
  Prostokaty     :ProstokatyTab;
   {Tablica prostokatow z ich listami sasiadow; prostokaty wystepuja w tej
    tablicy w porzadku zamykania.}
  PorzOtw        :BokiInfoTab;
   {Tu obliczymy permutacje numerow prostokatow z tablicy prostokaty,
    wyznaczajaca ich porzadek zgodny z kolejnoscia otwierania.}
  NrOtw,NrZam    :integer;
   {NrOtw -  numer ostatnio otwartego prostokata;
    NrZam -  numer ostatnio zamknietego prostokata.}
  Otwier         :boolean;
   {Zmienna otwier pozwoli klasyfikowac boki wielokata jako otwierajace
    lub zamykajace. Niech w^[i]--w^[i+1] bedzie bokiem poziomym z najnizszej
    warstwy. Oczywiscie ten bok jest otwierajacy. Wowczas wartoscia Otwier
    bedzie (w^[i] < w^[i+1]). Latwo zauwazyc, ze poziomy bok w^[j]--w^[j+1]
    jest otwierajacy wtedy i tylko wtedy, gdy Otwier = (w^[j] < w^[j+1]).
    Natomiast pionowy bok (w^[j]--w^[j+1]) jest otwierajacy wtedy i tylko
    wtedy, gdy Otwier = (w^[j] > w^[j+1]).}
  xWe,yWe,xSk,ySk:integer;
   {Wspolrzedne punktow, odpowiednio, wejsciowego i ukrycia skarbu.}

 procedure SortujBoki(var Boki:BokiInfoWsk; w1,w2:integer);
  {Procedura porzadkowania bokow.
   Dane:
    Boki[1..n/2] - numery bokow pionowych (poziomych) do posortowania;
    w1=1,2; w2=1,2; w1 <> w2; klucz pierwotny w sortowaniu - wsp.
    o numerze w1; klucz wtorny w sortowaniu - wsp. o numerze w2.
   Wynik:
    Boki[1..n/2] - permutacja numerow bokow pionowych (poziomych)
    wyznaczajaca ich kolejnosc w porzadku leksykograficznym wzgledem
    wspolrzednych w1 i w2 dolnych koncow (lewych koncow -- w przypadku
    bokow poziomych).
   Metoda: adaptacja algorytmu HeapSort.}

  function Mniejszy(v1,v2:WierzchTab):boolean;
   {Porownanie dwoch wierzcholkow, przy czym najpierw porownuje sie
    wspolrzedne w1, a nastepnie w2.}
  begin
    Mniejszy:=(v1[w1]<v2[w1]) or ((v1[w1]=v2[w1]) and (v1[w2]<v2[w2]))
  end; {Mniejszy}

  procedure ldKoniec(i:integer; var v:WierzchTab);
    {Procedura oblicza lewy (dolny) koniec odcinka poziomego (pionowego).}
   var i1 : integer;
  begin
    if i=n then i1:=1 else i1:=i+1;
    if Mniejszy(w^[i],w^[i1]) then v:=w^[i] else v:=w^[i1]
  end; {ldKoniec}

  procedure Zamien(i,j:integer);
    {Zamienia Boki^[i] z Boki^[j].}
   var b:integer;
  begin
    b:=Boki^[i];
    Boki^[i]:=Boki^[j];
    Boki^[j]:=b
  end; {Zamien}

  procedure PoprawKopiec(l,p:integer);
   {Dane: kopiec Boki[l+1..p];  Wynik: Kopiec[l..p]}
   var j:integer; v0,v1,v2:WierzchTab;
  begin
   ldKoniec(Boki^[l],v0);
   while 2*l<=p do begin
    j:=2*l;
    ldKoniec(Boki^[j],v1);
    if j<p then begin
     ldKoniec(Boki^[j+1],v2);
     if Mniejszy(v1,v2) then begin v1:=v2;  j:=j+1 end
    end; {if }
    if Mniejszy(v0,v1) then begin Zamien(l,j);  l:=j end
    else l:=p
   end {while }
  end; {PoprawKopiec}

   {Wlasciwe sortowanie}

  var i:integer;
 begin
  for i:=nDiv2 div 2 downto 1 do PoprawKopiec(i,nDiv2);
  for i := nDiv2 downto 2 do begin
   Zamien(1,i);
   PoprawKopiec(1,i-1)
  end
 end; {SortujBoki}

  {******************OPERACJE NA PRZEKROJU***************************}

 procedure IniPrzekroj;
   {Inicjalizacja Przekroju. ll - liczba lisci w przekroju.
    Przekroj[0] - straznik, zawsze rowny 0.}
  var i:integer;
 begin
  new(Przekroj);
  for i:=0 to 2*ll-1 do Przekroj^[i]:=0
 end; {IniPrzekroj}

 procedure WstawDoPrzekroju(b:integer);
   {Wstawienie boku poziomego w^[b]--w^[b+1] do przekroju.}
  var Poz:integer;
 begin
  Poz:=ll+Warstwy^[b]-1;
   {Poz jest nr. liscia w drzewie przekroju odpowiadajacem warstwie
    zawierajacej bok w^[b]--w^[b+1].}

   {Zeby wstawic bok b do Przekroju wystarczy wpisywac b do kolejnych
    wezlow na sciezce prowadzacej od liscia Przekroj^[Poz] do korzenia
    Przekroj^[1], az do napotkania pierwszego wezla, ktorego prawy brat
    ma wartosc > 0 lub osiagniecia korzenia.}
  repeat
   Przekroj^[Poz]:=b;
   if not odd(Poz) then
    {Jestesmy w wezle drzewa, ktory jest lewym synem swojego ojca.}
    if Przekroj^[Poz+1]>0 then b:=0;
   Poz:=Poz div 2
  until (b=0) or (Poz=0)
 end; {WstawDoPrzekroju}

 procedure UsunZPrzekroju(b:integer);
   {Usuniecie z przekroju boku w^[b]--w^[b+1].
    Operacja odwrotna do wstawiania.}
  var b1,Poz:integer;
 begin
  Poz:=ll+Warstwy^[b]-1;
   {Startujemy od liscia zawierajacego bok, ktory chcemy usnac z przekroju.}

   {Znajdz pierwszy wezel na sciezce prowdzacej od liscia Przekroj^[Poz]
    do korzenia Przekroj^[1], ktorego wartosc jest rozna od b, lub lewy
    brat ma wartosc rozna od zera.}
  repeat
   Przekroj^[Poz]:=0;
   Poz:=Poz div 2
  until (Przekroj^[2*Poz]<>0) or (Przekroj^[Poz]<>b);
  if Przekroj^[2*Poz]<>0 then begin
   {Jezeli Przekroj^[2*Poz]<>0 to na sciezce od Przekroj^[Poz] do korzenia
    we wszystkich wezlach, w ktorych bylo zapisane b wpisz b1.}
   b1:=Przekroj^[2*Poz];
   repeat
    Przekroj^[Poz]:=b1;
    Poz:=Poz div 2
   until Przekroj^[Poz]<>b
  end {if }
 end; {UsunZPrzekroju}

 function Nast(b:integer):integer;
   {wartoscia jest nr boku z pierwszej niepustej warstwy w przekroju
    polozonej na prawo od warstwy zawierajacej bok w^[b]--w^[b+1];
    funkcja jest wywolywana zawsze w przypadku, gdy taka warstwa istnieje.}
  var Poz,zPrawej:integer;
 begin
  Poz:=ll+Warstwy^[b]-1;

  {Znajdz w drzewie najmniejszego przodka liscia Przekroj^[Poz], w ktorego
   lewym poddrzewie jest lisc Przekroj^[Poz], a w prawym niepusta warstwa.}

  zPrawej:=0;
  repeat
   if not odd(Poz) then zPrawej:=Przekroj^[Poz+1];
   Poz:=Poz div 2
  until zPrawej>0;
   {Przejdz do prawego poddrzewa}
  Poz:=2*Poz+1;

   {Znajdz skrajnie lewy lisc w poddrzewie o korzeniu Przekroj^[Poz] zawierajacy
    niepusta wartswe. Pamietaj, ze liscie maja numery od ll do 2ll-1.}

  while 2*Poz<=2*ll-1 do
   {Jesli w lewym poddrzewie jest niepusta warstwa to zejdz do lewego
    poddrzewa, w przeciwnym razie zejdz do prawego poddrzewa.}
   if Przekroj^[2*Poz]>0 then Poz:=2*Poz else Poz:=2*Poz+1;
  Nast:=Przekroj^[Poz]
 end; {Nast}

 function Poprz(b:integer):integer;
   {Wartoscia jest bok z pierwszej niepustej warstwy w przekroju
    polozonej na lewo od wartswy zawierajacej bok w^[b]--w^[b+1].
    Funkcja jest wywolywana zawsze w przypadku, gdy taka warstwa istnieje.}
  var Poz,zLewej:integer;
 begin
  Poz:=ll+Warstwy^[b]-1;

   {Znajdz w drzewie najmniejszego przodka liscia Przekroj^[Poz], w ktorego
    prawym poddrzewie jest lisc Przekroj^[Poz], a w lewym niepusta warstwa.
    Obliczenia sa podobne do tych z funkcji Nast. }

  zLewej:=0;
  repeat
   if odd(Poz) then zLewej:=Przekroj^[Poz-1];
   Poz:=Poz div 2
  until zLewej>0;
  Poz:=2*Poz;
  while 2*Poz+1<=2*ll-1 do
   if Przekroj^[2*Poz+1]>0 then Poz:=2*Poz+1 else Poz:=2*Poz;
  Poprz:=Przekroj^[Poz]
 end; {Poprz}

 function JestWPrzekroju(b:integer):boolean;
   {Wartoscia jest true wtedy i tylko wtedy, gdy bok w^[b]--w^[b+1]
    jest w przekroju.}
 begin
  JestWPrzekroju:=Przekroj^[ll+Warstwy^[b]-1]=b
 end; {JestWPrzekroju}

  {******************OPERACJE NA PRZEKROJU***************************}

  {******************PROCEDURY I FUNKCJE WLASCIWEGO PROGRAMU**********}

 procedure Inicjalizacja;
   {Wczytanie danych wejsciowych i inicjalizacja struktur danych.}
  var
   f    :text;
   i,j,k:integer;
   u,v  :WierzchTab;
 begin
  assign(f,'ZAM.IN');
  reset(f);
  readln(f,n);
  nDiv2:=n div 2;
   {Wczytanie wierzcholkow (bokow) wielokata, pierwszym bokiem
    (w^[1]---w^[2]) bedzie bok poziomy. Jednoczesnie wyznaczamy jeden
    bok poziomy z najnizszej wrastwy.}
  new(w);
  readln(f,u[1],u[2]);
  readln(f,v[1],v[2]);
  if u[2]=v[2] then begin
   w^[1]:=u;  w^[2]:=v;  j:=1;
   for i:=3 to n do begin
    readln(f,w^[i][1],w^[i][2]);
    if w^[i][2]<w^[j][2] then j:=i
   end
  end {if }
  else begin
   w^[1]:=v;  j:=1;
   for i:=2 to n-1 do begin
    readln(f,w^[i][1],w^[i][2]);
    if w^[i][2]<w^[j][2] then j:=i
   end;
   w^[n]:=u
  end; {else }
   {w^[j]--w^[j+1] jest bokiem poziomym z najnizszej warstwie}
   {Wczytanie wspolrzednych punktow wejscia do zamku i ukrycia skarbu.}
  readln(f,xWe,yWe,xSk,ySk);
  close(f);
   {Przygotowanie testu na sprawdzanie statusu boku: otwierajacy-zamykajacy.}
  Otwier:=w^[j][1]<w^[j+1][1];
   {Po inicjalizacji
     w        : 1    2    3    4   ...    n-1     n
     Wierzch  : v_1  v_2  v_3  v_4        v_(n-1) v_n
     Bok      :    --   |   --    |   --   |     |

     v_1,v_2,...,v_n -  kolejnosc wystepowania wierzcholkow na obwodzie w;
     dla kazdego nieparzystego i, 1 <= i <n, w^[i] -- w^[i+1] jest
     otwierajacy wtedy i tylko wtedy, gdy (w^[i][1] < w^[i+1][1]) = Otwier;
     dla kazdego parzystego i, 1 < i<= n, w^[i] -- w^[i+1] jest otwierajacy
     wtedy i tylko wtedy, gdy (w^[i][2] > w^[i+1][2]) = Otwier.          }

   {Sortowanie bokow poziomych leksykograficznie wzgledem par (2-wsp,1-wsp.
    lewego konca, czyli w porzadku < zdefiniowanym w opisie algorytmu.
    Tutaj tablica BokiPionowe jest pomocnicza. Naszym celem jest obliczenie
    dla kazdego boku poziomego numeru warstwy, do ktorej ten bok nalezy.
    Policzymy takze liczbe wszystkich warstw.}

  new(BokiPionowe);
  for i:=1 to nDiv2 do BokiPionowe^[i]:=2*i-1;
  SortujBoki(BokiPionowe,2,1);
  new(Warstwy);
  lw:=0;  j:=-1;  ll:=1;
  for i:=1 to nDiv2 do begin
   k:=BokiPionowe^[i];
   if w^[k][2]>j then begin
    lw:=lw+1;  j:=w^[k][2];
    if lw>ll then ll:=2*ll
   end;
   Warstwy^[k]:=lw;  Warstwy^[k+1]:=lw
  end; {for i }

   {Sortowanie bokow pionowych leksykograficznie wzgledem dolnych koncow,
    czyli w porzadku < zdefiniowanym w opisie algorytmu.}

  for i:=1 to nDiv2 do BokiPionowe^[i]:=2*i;
  SortujBoki(BokiPionowe,1,2);

   {Inicjalizacja zmiennych zliczajacych ilosci otwieranych i zamykanych
    prostokatow.}
  NrOtw:=0;  NrZam:=0;
   {Inicjalizacja przekroju.}
  IniPrzekroj
 end; {Inicjalizacja}

 function Otwierajacy(i:integer):boolean;
   {Wartoscia jest true wtedy i tylko wtedy, gdy bok w^[i]--w^[i+1] jest
    otwierajacy. Dla i=n badamy bok w^[n]--w^[1].}
  var i1:integer;
 begin
  if i=n then i1:=1 else i1:=i+1;
  Otwierajacy:=Otwier=((w^[i][2]=w^[i1][2]) and (w^[i][1]<w^[i1][1]))
                      or
                      ((w^[i][1]=w^[i1][1]) and (w^[i][2]>w^[i1][2]))
 end; {Otwierajacy}

 procedure Sasiedzi(i:integer; var Gorny,Dolny:integer);
   {Procedura znajduje gornego i dolnego sasiada pionowego boku w^[i]--w^[i+1].
    Dla i=n badamy bok w^[n]--w^[1].}
  var i1:integer;
 begin
  if i=n then i1:=1 else i1:=i+1;
  if w^[i][2]<w^[i1][2] then begin Dolny:=i-1;  Gorny:=i1 end
  else begin Gorny:=i-1;  Dolny:=i1 end
 end; {Sasiedzi}

 procedure Konce(i:integer; var Lewy,Prawy:integer);
  {Procedure ustala lewy i prawy koniec poziomego boku w^[i]--w^[i+1].}
 begin
  if w^[i][1]<w^[i+1][1] then begin Lewy:=i;  Prawy:=i+1 end
  else begin Lewy:=i+1;  Prawy:=i end
 end; {Konce}

 procedure OtworzProstokat(b,l:integer);
   {Otwarcie prostokata ograniczonego z dolu przez bok poziomy
    w^[b]--w^[b+1] i z lewej strony przez linie pionowa o numerze l.}
 begin
  LewyBok[b div 2+1]:=l;
  NrOtw:=NrOtw+1;
  Otw[b div 2+1]:=NrOtw
 end; {OtworzProstokat}

 procedure ZamknijProstokat(e,f,p:integer);
   {Procedura zamyka prostokat ograniczony z dolu przez bok w^[e]--w^[e+1],
    z gory przez bok w^[f]--w^[f+1] i z prawym bokiem zawartym w linii o
    numerze p.}
 begin
  NrZam:=NrZam+1;
  with Prostokaty[NrZam] do begin
   ld[1]:=LewyBok[e div 2+1];  ld[2]:=w^[e][2];
   PorzOtw[Otw[e div 2+1]]:=NrZam;
   Wys:=w^[f][2]-ld[2];
   Szer:=p-ld[1];
   ListaSasiednichPr:=nil
  end {with }
 end; {ZamknijProstokat}

 procedure KolejnyBokPionowy(b : integer);
   {Przetwarzanie kolejnego boku pionowego w^[b]--w^[b+1].}
  var g,d,a,c,p,lk,pk:integer;
 begin
  p:=w^[b][1];
   {p jest numerem linii pionowej zawierajacej bok pionowy w^[b]--w^[b+1].}
  Sasiedzi(b,g,d);
   {w^[g]--w^[g+1] - gorny bok poziomy sasiadujacy z w^[i]--w^[i+1];
    w^[d]--w^[d+1] - dolny bok poziomy sasiadujacy z w^[i]--w^[i+1].}
  if Otwierajacy(b) then
   case Otwierajacy(d) of
    true: case Otwierajacy(g) of
           true:  { ___     przypadek (a)
                       |___                }
                 begin
                  a:=Nast(g);
                  ZamknijProstokat(g,a,p);
                  Konce(a,lk,pk);
                   {w^[lk], w^[pk] -- odpowiednio, lewy i prawy koniec w^[a]--w^[a+1]}
                  if w^[pk][1] = p then UsunZPrzekroju(a);
                  UsunZPrzekroju(g);
                  WstawDoPrzekroju(d);
                  OtworzProstokat(d,p)
                 end; {przypadek (a)}
           false:  {  ___
                     |___     przypadek (b) }
                 begin
                  WstawDoPrzekroju(d);
                  WstawDoPrzekroju(g);
                  OtworzProstokat(d,p)
                 end {przypadek (b)}
          end; {case Otwierajacy(g)}
    false:case Otwierajacy(g) of
           true:  {___
                   ___|       przypadek (c) }
                 begin
                  if JestWPrzekroju(d) then begin
                    {Jesli w^[b]--w^[b+1] jest zamykajacy dla
                     prostokata ograniczonego z gory przez w^[d]--w^[d+1].}
                   a:=Poprz(d);
                   ZamknijProstokat(a,d,p);
                   UsunZPrzekroju(d);
                   OtworzProstokat(a,p)
                  end; {if }
                  a:=Nast(g);
                  ZamknijProstokat(g,a,p);
                  Konce(a,lk,pk);
                   {w^[lk], w^[pk] -- odpowiednio,
                    lewy i prawy koniec w^[a]--w^[a+1]}
                  if w^[pk][1]=p then UsunZPrzekroju(a);
                  UsunZPrzekroju(g)
                 end; {case Otwierajacy(g)}
           false:  {        ___
                        ___|        przypadek (d) }
                 begin
                  if JestWPrzekroju(d) then begin
                    {Jesli w^[b]--w^[b+1] jest zamykajacy dla
                     prostokata ograniczonego z gory przez w^[d]--w^[d+1].}
                   a:=Poprz(d);
                   ZamknijProstokat(a,d,p);
                   UsunZPrzekroju(d);
                   OtworzProstokat(a,p)
                  end; {if }
                  WstawDoPrzekroju(g)
                 end {przypadek (d)}
          end {case Otwierajacy(g)}
   end {case Otwierajacy(d)}
  else {if not Otwierajacy(b)}
   case Otwierajacy(d) of
    true: case Otwierajacy(g) of
           true:  {       ___
                      ___|       przypadek (e) }
                 begin
                  a:=Nast(d);
                  ZamknijProstokat(d,a,p);
                  Konce(a,lk,pk);
                  if w^[pk][1]=p then UsunZPrzekroju(a);
                  UsunZPrzekroju(d);
                  OtworzProstokat(g,p);
                  WstawDoPrzekroju(g)
                 end; {przypadek (e)}
           false:   {  ___
                       ___|    przypadek (f) }
                 begin
                  ZamknijProstokat(d,g,p);
                  UsunZPrzekroju(d);
                  UsunZPrzekroju(g)
                 end {przypadek(f)}
          end; {case Otwierajacy(g)}
    false:case Otwierajacy(g) of
           true:  {  ___
                    |___    przypadek (g) }
                 begin
                  a:=Poprz(d);
                  if LewyBok[a div 2+1]<p then begin
                    {w^[b]--w^[b+1] jest zamykajacy dla prostokata z lewej strony.}
                   c:=Nast(d);
                   ZamknijProstokat(a,c,p);
                   OtworzProstokat(a,p);
                   Konce(c,lk,pk);
                   if w^[pk][1]=p then UsunZPrzekroju(c)
                  end; {if }
                  OtworzProstokat(g,p);
                  WstawDoPrzekroju(d);
                  WstawDoPrzekroju(g)
                 end; {przypadek (g)}
           false:   { ___
                         |___  przypadek (h) }
                 begin
                  a:=Poprz(g);
                  if LewyBok[a div 2+1]<p then begin
                    {w^[b]--w^[b+1] jest zamykajacy dla prostokata z lewej strony.}
                   ZamknijProstokat(a,g,p);
                   OtworzProstokat(a,p);
                   UsunZPrzekroju(g)
                  end; {if }
                  WstawDoPrzekroju(d)
                 end {przypadek (h)}
          end {case Otwierajacy(g)}
   end {case Otwierajacy(d)}
 end; {KolejnyBokPionowy}

 procedure DzielNaProstokaty;
  var b:integer;
 begin
  for b:=1 to nDiv2 do KolejnyBokPionowy(BokiPionowe^[b]);
  dispose(Przekroj)   {tablica Przekroj nie bedzie wiecej potrzebna}
 end; {DzielNaProstokaty}

 procedure BudujDrzewoProstokatow;
  var
   v  :WskDoNastProstokataNaLiscie;
   i,j:integer;
 begin
  i:=1;  j:=1;
  while (i<=NrZam) and (j<=NrOtw) do
   with Prostokaty[i] do
    if ((ld[1]+Szer)<Prostokaty[PorzOtw[j]].ld[1])
       or (((ld[1]+Szer)=Prostokaty[PorzOtw[j]].ld[1])
       and (ld[2]+Wys<=Prostokaty[PorzOtw[j]].ld[2]))
    then i:=i+1
    else
     if ((ld[1]+Szer)=Prostokaty[PorzOtw[j]].ld[1])
        and (ld[2]<(Prostokaty[PorzOtw[j]].ld[2]+Prostokaty[PorzOtw[j]].Wys))
     then begin
      new(v);
      v^.nr:=PorzOtw[j];
      v^.nast:=ListaSasiednichPr;
      ListaSasiednichPr:=v;
      new(v);
      v^.nr:=i;
      v^.nast:=Prostokaty[PorzOtw[j]].ListaSasiednichPr;
      Prostokaty[PorzOtw[j]].ListaSasiednichPr:=v;
      if ld[2]+Wys>Prostokaty[PorzOtw[j]].ld[2]+Prostokaty[PorzOtw[j]].Wys
      then j:=j+1
      else
       if ld[2]+Wys<Prostokaty[PorzOtw[j]].ld[2]+Prostokaty[PorzOtw[j]].Wys
       then i:=i+1
       else begin i:=i+1;  j:=j+1 end
     end {then }
     else j:=j+1
 end; {BudujDrzewoProstokatow}

 function ZnajdzProstokat(x,y:integer):integer;
   {Znajduje prostokat zawierajacy punkt (x,y).}
  var
   Jest:boolean;
   i   :integer;
 begin
  Jest:=false;  i:=1;
  while not Jest do
   with Prostokaty[i] do
    if (x>=ld[1]) and (x<=(ld[1]+Szer)) and
       (y>=ld[2]) and ((y<=ld[2]+Wys))
    then Jest:=true
    else i:=i+1;
  ZnajdzProstokat:=i
 end; {ZnajdzProstokat}

 function Odleglosc(pr1,pr2:ProstokatTab; var x,y:integer):integer;
   {Dane: sasiadujace prostokaty pr1, pr2 oraz punk (x,y) z pr1.
    Wynik: odleglosc miedzy (x,y) i punktem (a,b) z pr2 polozonym najblizej
    (x,y); punkt (a,b).}
  var a,b,b1,b2:integer;
 begin
  if pr1.ld[1]<pr2.ld[1] then a:=pr2.ld[1]
  else a:=pr1.ld[1];
  if pr1.ld[2]<=pr2.ld[2] then b1:=pr2.ld[2] else b1:=pr1.ld[2];
  if pr1.ld[2]+pr1.Wys<=pr2.ld[2]+pr2.Wys
  then b2:=pr1.ld[2]+pr1.Wys
  else b2:=pr2.ld[2]+pr2.Wys;
  if y>b2 then b:=b2
  else
   if y<b1 then b:=b1
   else b:=y;
  Odleglosc:=abs(a-x)+abs(b-y);
  x:=a;  y:=b
 end; {Odleglosc}

 function NajkrotszaOdleglosc:longint;
   {Obliczanie dlugosci drogi optymalnej miedzy punktem startowym, a punktem
    ukrycia skarbu.}
  var
   Droga             :array[0..MaksNDiv2] of integer;
   i,j,Start,Meta,x,y:integer;
   Odl               :longint;
   Jest              :boolean;
 begin
   {Znajdowanie ciagu prostokatow przez ktore przechodzi kazda optymalna droga.}
  Droga[0]:=0;
  Start:=znajdzProstokat(xWe,yWe);
  Meta:=znajdzProstokat(xSk,ySk);
  Droga[1]:=Start;  i:=1;
  while Droga[i]<>Meta do begin
   with Prostokaty[Droga[i]] do
    if ListaSasiednichPr=nil then i:=i-1
    else begin
     if ListaSasiednichPr^.nr<>Droga[i-1] then begin
      i:=i+1;  Droga[i]:=ListaSasiednichPr^.nr
     end; {if }
     ListaSasiednichPr:=ListaSasiednichPr^.nast
    end {else }
  end; {while}
   {Obliczanie dlugosci optymalnej drogi zgodnie ze wzorem (1) z opisu.}
  x:=xWe;  y:=yWe;  Odl:=0;
  for j:=1 to i-1 do
   Odl:=Odl+Odleglosc(Prostokaty[Droga[j]],Prostokaty[Droga[j+1]],x,y);
  NajkrotszaOdleglosc:=Odl+abs(x-xSk)+abs(y-ySk)
 end; {NajkrotszaOdleglosc}

 procedure WypiszWynik;
  var
   f:text;
   i:integer;
   v:WskDoNastProstokataNaLiscie;
 begin
  assign(f,'ZAM.OUT');  rewrite(f);
  writeln(f,NajkrotszaOdleglosc);
  close(f)
 end; {WypiszWynik}

begin {Program glowny}
 Inicjalizacja;
 DzielNaProstokaty;
 BudujDrzewoProstokatow;
 WypiszWynik
end. {Zamek}